
/*----------------------------------------------------------------------------+
 |                                                                            |
 |                  Oregon Sensor Support for HEYU                            |
 |                Copyright 2008 Charles W. Sullivan                          |
 |                      All Rights Reserved                                   |
 |                                                                            |
 |                                                                            |
 | This software is licensed free of charge for non-commercial distribution   |
 | and for personal and internal business use only.  Inclusion of this        |
 | software or any part thereof in a commercial product is prohibited         |
 | without the prior written permission of the author.  You may copy, use,    |
 | and distribute this software subject to the following restrictions:        |
 |                                                                            |
 |  1)	You may not charge money for it.                                      |
 |  2)	You may not remove or alter this license, copyright notice, or the    |
 |      included disclaimers.                                                 |
 |  3)	You may not claim you wrote it.                                       |
 |  4)	If you make improvements (or other changes), you are requested        |
 |	to send them to the Heyu maintainer so there's a focal point for      |
 |      distributing improved versions.                                       |
 |                                                                            |
 | As used herein, HEYU is a trademark of Daniel B. Suthers.                  | 
 | X10, CM11A, and ActiveHome are trademarks of X-10 (USA) Inc.               |
 | The author is not affiliated with either entity.                           |
 |                                                                            |
 | Charles W. Sullivan                                                        |
 | Co-author and Maintainer                                                   |
 | Greensboro, North Carolina                                                 |
 | Email ID: cwsulliv01                                                       |
 | Email domain: -at- heyu -dot- org                                          |
 |                                                                            |
 | Disclaimers:                                                               |
 | THERE IS NO ASSURANCE THAT THIS SOFTWARE IS FREE OF DEFECTS AND IT MUST    |
 | NOT BE USED IN ANY SITUATION WHERE THERE IS ANY CHANCE THAT ITS            |
 | PERFORMANCE OR FAILURE TO PERFORM AS EXPECTED COULD RESULT IN LOSS OF      |
 | LIFE, INJURY TO PERSONS OR PROPERTY, FINANCIAL LOSS, OR LEGAL LIABILITY.   |
 |                                                                            |
 | TO THE EXTENT ALLOWED BY APPLICABLE LAW, THIS SOFTWARE IS PROVIDED "AS IS",|
 | WITH NO EXPRESS OR IMPLIED WARRANTY, INCLUDING, BUT NOT LIMITED TO, THE    |
 | IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.|
 |                                                                            |
 | IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL THE AUTHOR BE LIABLE    |
 | FOR DAMAGES, INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL   |
 | DAMAGES ARISING OUT OF THE USE OR INABILITY TO USE THIS SOFTWARE EVEN IF   |
 | THE AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.            |
 |                                                                            |
 +----------------------------------------------------------------------------*/



#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <syslog.h>
#include <ctype.h>
#include <time.h>
#if defined(SYSV) || defined(FREEBSD) || defined(OPENBSD)
#include <string.h>
#else
#include <strings.h>
#endif

#include "x10.h"
#include "process.h"
#include "x10state.h"
#include "oregon.h"

extern int sptty;
extern int i_am_state, i_am_aux, i_am_relay, i_am_rfxmeter;
extern char *funclabel[];
extern void create_flagslist ( unsigned char, unsigned long, char * );
extern double celsius2temp ( double, char, double );
extern double temp2celsius ( double, char, double );
extern void intvstrf ( char *, time_t, time_t );

extern CONFIG config;
extern CONFIG *configp;

extern struct x10global_st x10global;
extern x10hcode_t *x10state;

extern unsigned int modmask[NumModMasks][16];
extern int lock_for_write(), munlock();

extern unsigned short ore_ignore[];
extern int            ore_ignore_size;

#if 0
int str2hex ( char *str, unsigned char *buf, int *bufsiz )
{
   char minibuf[16];
   char *sp;
   int  j;
   
   if ( (strlen(str) % 2) != 0 ) {
      fprintf(stderr, "Invalid string length %d\n", strlen(str));
      return 1;
   }

   sp = str;

   j = 0;
   while ( *sp != '\0' ) {
      minibuf[0] = *sp++;
      minibuf[1] = *sp++;
      minibuf[2] = '\0';
      buf[j++] = (unsigned char)strtol(minibuf, NULL, 16);
   }
   if ( buf[0] != 8 * (j - 1) ) {
      fprintf(stderr, "Miscount: buf[0] = %d, bufsiz = %d\n", buf[0]/8, j - 1);
      return 1;
   }

   memmove(buf, buf + 1, j - 1);
   *bufsiz = j - 1;

   return 0;
}

unsigned int bcd2dec ( unsigned int bcd, int digits )
{
   unsigned int dec = 0, mult = 1, mask =0x0f;
   int          j;

   for ( j = 0; j < digits; j++ ) {
      dec += mult * ((bcd & mask) >> (4 * j));
      mult *= 10;
      mask <<= 4;
   }
   return dec;
}
#endif


int is_ore_ignored ( unsigned short saddr )
{
#ifdef HASORE
   int j = 0;

   while ( j < ore_ignore_size ) {
      if ( saddr == ore_ignore[j] )
         return 1;
      j++;
   }
#endif
   return 0;
}

#ifdef HASORE
unsigned char nybble_sum( int nbytes, unsigned char *buf )
{
   int   j;
   unsigned char sum = 0;

   for ( j = 0; j < nbytes; j++ ) {
      sum += (buf[j] & 0x0f) + ((buf[j] >> 4) & 0x0f);
   }
   return sum;
}
      
unsigned char cse ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(7, buf) - buf[7];
   
   return (unsigned char)(ssum & 0xff) - 0x18u;
}   

unsigned char cs7 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(7, buf) - buf[7];
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}   

unsigned char cs8 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(8, buf) - buf[8];
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char cs9 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(9, buf) - buf[9];
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}   

unsigned char cs10 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(10, buf) - buf[10] ;
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char cs11 ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(11, buf) - buf[11] ;
   
   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char cs2HL ( unsigned char *buf )
{
  short ssum;

  ssum = nybble_sum(8, buf) + (buf[8] & 0x0f) -
        ((buf[8] >> 4) & 0x0f) - ((buf[9] << 4) & 0xf0);

  return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char csRHL ( unsigned char *buf )
{
  short ssum;

  ssum = nybble_sum(9, buf) + (buf[9] & 0x0f)
         - ((buf[9] >> 4) & 0x0f) - ((buf[10] << 4) & 0xf0);
  
  return (unsigned char)(ssum & 0xff) - 0x0A;
}

unsigned char csWHL ( unsigned char *buf )
{
   short ssum;

   ssum = nybble_sum(6, buf) + (buf[6] & 0x0f) - 
          ((buf[6] >> 4) & 0x0f) - ((buf[7] << 4) & 0xf0);

   return (unsigned char)(ssum & 0xff) - 0x0Au;
}

unsigned char csWgt1 ( unsigned char *buf )
{
   if ( (buf[0] & 0xf0) != (buf[5] & 0xf0) ||
        (buf[1] & 0x0f) != (buf[6] & 0x0f) ||
        (buf[6] & 0xf0) != 0x10            ||
        (buf[5] & 0x0e) != 0x00               )
      return 1;
   return 0;
}

unsigned char csNull ( unsigned char *buf )
{
  return 0;
}

unsigned char csFail ( unsigned char *buf )
{
  return 1;
}

/* Types
GR101:    GR101
TEMP1:    THR128,THx138
TEMP2:    THN132N,THWR288
TH1:      THGN122N,THGR122NX,THGR228N,THGR268
TH2:      THGR810
TH3:      RTGR328N
TH4:      THGR328 (WIP)
TH5:      WTGR800
TH6:      THGR918
THB1:     BTHR918
THB2:     BTHR918N,BTHR968
RAIN1:    RGR126,RGR682,RGR918
RAIN2:    PCR800
RAIN3:
WIND1:    WTGR800
WIND2:    WGR800
WIND3:    STR918,WGR918
UV1:      UVR138
UV2:      UVN800
DT1:      RTGR328N
ELEC1:    cent-a-meter
WEIGHT1:  BWR101,BWR102
*/


static struct oregon_st {
   unsigned char  minbits;
   unsigned char  trulen;
   unsigned char  id0mask;
   unsigned char  id1mask;
   unsigned char  id0;
   unsigned char  id1;
   int            nvar;
   unsigned char  chsize;
   unsigned char  batmode;
   unsigned short bpoffset;
   unsigned char  support_mode;  /* 0 = Unsupported; 1 = Normal sensor; 2 = Dummy sensor */
   unsigned char  addr_mask;
   int            addr_byte;
   char           *slabel;
   unsigned char  subtype;
   unsigned char  (*chksumf)(unsigned char *);
} orechk[] = {
   {88,  11, 0xff, 0xff, 0x5a, 0x5d, 3, 3, 1, 795, 1, 0xff, 4, "THB1", OreTHB1,    cs10  },
   {88,  11, 0xff, 0xff, 0x5a, 0x6d, 3, 3, 2, 856, 1, 0xff, 4, "THB2", OreTHB2,    cs10  },
   {72,   9, 0xff, 0xff, 0x0a, 0x4d, 1, 3, 1, 0,   1, 0xff, 4, "T1",   OreTemp1,   cs8   },
   {72,   9, 0xff, 0xff, 0x1a, 0x2d, 2, 3, 1, 0,   1, 0xff, 4, "TH1",  OreTH1,     cs8   },
   {72,   9, 0xff, 0xff, 0xfa, 0x28, 2, 4, 1, 0,   1, 0xff, 4, "TH2",  OreTH2,     cs8   },
   {72,   9, 0x0f, 0xff, 0x0a, 0xcc, 2, 4, 1, 0,   1, 0xff, 4, "TH3",  OreTH3,     cs8   },
   {72,   9, 0x0f, 0xff, 0xca, 0x2c, 2, 3, 1, 0,   1, 0xff, 4, "TH4",  OreTH4,     cs8   },
   {72,   9, 0xff, 0xff, 0xfa, 0xb8, 2, 0, 2, 0,   1, 0xff, 4, "TH5",  OreTH5,     cs8   },
   {72,   9, 0xff, 0xff, 0x1a, 0x3d, 2, 3, 2, 0,   1, 0xff, 4, "TH6",  OreTH6,     cs8   },
   {60,   8, 0xff, 0xff, 0xea, 0x4c, 1, 3, 1, 0,   1, 0xff, 4, "T2",   OreTemp2,   csWHL },
   {56,   7, 0x0f, 0x00, 0x0C, 0x00, 1, 0, 0, 0,   1, 0xf0, 2, "WGT1", OreWeight1, csWgt1},

   {88,  11, 0xff, 0xff, 0xf0, 0xf0, 1, 0, 0, 0,   2, 0xff, 4, "TEMU",   OreTemu,   csFail}, /* Dummy */
   {88,  11, 0xff, 0xff, 0xf0, 0xf0, 2, 0, 0, 0,   2, 0xff, 4, "THEMU",  OreTHemu,  csFail}, /* Dummy */
   {88,  11, 0xff, 0xff, 0xf0, 0xf0, 3, 0, 0, 0,   2, 0xff, 4, "THBEMU", OreTHBemu, csFail}, /* Dummy */

   /* Unsupported Oregon sensors */

   {96,  12, 0xff, 0xff, 0x8a, 0xec, 2, 4, 2, 0,   0, 0xff, 4, "DT1",  OreDT1,     cs11  },
   {84,  11, 0xff, 0xff, 0x2a, 0x19, 2, 0, 2, 0,   0, 0xff, 4, "R2",   OreRain2,   csRHL },
   {84,  11, 0xff, 0xff, 0x06, 0xe4, 2, 3, 1, 0,   0, 0xff, 4, "R3",   OreRain3,   csRHL },
   {80,  10, 0xff, 0xff, 0x1a, 0x99, 2, 0, 2, 0,   0, 0xff, 4, "WS1",  OreWind1,   cs9   },
   {80,  10, 0xff, 0xff, 0x1a, 0x89, 2, 0, 2, 0,   0, 0xff, 4, "WS2",  OreWind2,   cs9   },
   {80,  10, 0xff, 0xff, 0x3a, 0x0d, 3, 0, 2, 0,   0, 0xff, 4, "WS3",  OreWind3,   cs9   },
   {80,  10, 0xff, 0xff, 0x2a, 0x1d, 2, 0, 2, 0,   0, 0xff, 4, "R1",   OreRain1,   cs2HL },
   {64,   8, 0xff, 0xff, 0xda, 0x78, 1, 0, 1, 0,   0, 0xff, 4, "UV2",  OreUV2,     cs7   },
   {64,   8, 0xff, 0xf0, 0xea, 0x00, 3, 0, 0, 0,   0, 0xff, 4, "EL1",  OreElec1,   cse   },
   {64,   8, 0xff, 0xf0, 0xea, 0x20, 3, 0, 0, 0,   0, 0xff, 4, "EL1",  OreElec1,   cse   },
   {64,   8, 0x0f, 0x00, 0x03, 0x00, 2, 0, 0, 0,   0, 0xf0, 2, "WGT2", OreWeight2, csNull},    /* OreGR101 */
   {60,   8, 0xff, 0xff, 0xea, 0x7c, 1, 0, 1, 0,   0, 0xff, 4, "UV1",  OreUV1,     csWHL },

};
int orechk_size = sizeof(orechk)/sizeof(struct oregon_st);


unsigned char channelval( char *chstr, unsigned char chsize,
                                       unsigned char channel )
{
   unsigned char chval;

   switch ( chsize ) {
      case 3 :
         channel &= 0x07;
         switch ( channel ) {
            case 0 :
            case 1 :  chval = 1; sprintf(chstr, "Ch 1 "); break;
            case 2 :  chval = 2; sprintf(chstr, "Ch 2 "); break;
            case 4 :  chval = 3; sprintf(chstr, "Ch 3 "); break;
            default : chval = 0; sprintf(chstr, "Ch 0 "); break;
         }
         break;
      case 4 :
         chval = channel;
         sprintf(chstr, "Ch %d ", channel);
         break;
      case 0 :
         chval = 1;
         sprintf(chstr, "Ch 1 ");
         break;
      default :
         chval = 0;
         *chstr = '\0';
         break;
   }
   return chval;
}

#endif /* HASORE */

 
unsigned char oretype ( unsigned char *xbuf, unsigned char *subindx,
                        unsigned char *trulen, unsigned short *addr, int *nseq )
{
#ifdef HASORE
   int j = 0;

   while ( j < orechk_size ) {
      if ( (xbuf[0] >= orechk[j].minbits) &&
           (xbuf[1] & orechk[j].id0mask) == orechk[j].id0 &&
           (xbuf[2] & orechk[j].id1mask) == orechk[j].id1 &&
           orechk[j].chksumf(xbuf + 1) == 0 ) {
         *trulen = orechk[j].trulen;
         *addr = (unsigned short)(xbuf[orechk[j].addr_byte] & orechk[j].addr_mask);
         *nseq = (int)orechk[j].nvar;
         *subindx = j;
         if ( orechk[j].support_mode || configp->disp_ore_all == YES ) {
            return 0;
         }
         else {
            return 1;
         }
      }
      j++;
   }
   return 1;
#else
   return 1;
#endif /* HASORE */
}


/*----------------------------------------------------------------------------+
 | Translate Oregon RF data                                                   |
 | Buffer reference: ST_COMMAND, ST_GENLONG, plus:                            |
 |  vtype, subindx, vidhi, vidlo, seq, buflen, buffer                         |
 |   [2]     [3]     [4]    [5]   [6]    [7]    [8+]                          |
 +----------------------------------------------------------------------------*/
char *translate_oregon( unsigned char *buf, unsigned char *sunchanged, int *launchp )
{
#ifdef HASORE
   static char    outbuf[160];
   static char    interval[32];
   char           flagslist[80];
   ALIAS          *aliasp;
   LAUNCHER       *launcherp;
   extern unsigned int signal_source;
   unsigned char  func, *vdatap, vtype, subindx, subtype, seq, channel_code;
   int            humid, lasthumid;
   int            baro, lastbaro;
   unsigned short vident, idmask;
   int            unit, loc;
   int            dtempc, lastdtempc, delta, nvar;
   int            cweight, lastcweight;
   int            buflen;
   unsigned long  longvdata;

   char           hc;

   unsigned char  actfunc, xactfunc;
   unsigned int   trigaddr, mask, active, trigactive;
   unsigned int   bmaplaunch;
   unsigned long  afuncmap, xfuncmap;

   int            j, k, found, index = -1;
   unsigned char  hcode, ucode, trig = 0;
   unsigned int   bitmap = 0;
   unsigned long  vflags = 0;

   static unsigned long prevdata[3], lastdata[3];
   static char          chstr[8];
   static unsigned char channel;
   static unsigned char blevel, wgtnum;
   static unsigned char status;
   static char          batstr[32];
   unsigned short uknbytes;

   double temp, dbaro, weight;

   static char *typename[] = {"Std", "Ent", "Sec", "RFXSensor", "RFXMeter",
     "?", "??", "???", "????", "DigiMax", "Oregon", "Noise", "Noise2"};

   launcherp = configp->launcherp;
   aliasp = configp->aliasp;

   *launchp = -1;
   *sunchanged = 0;
   flagslist[0] = '\0';
   *outbuf = '\0';


   vtype   = buf[2];
   subindx = buf[3];
   seq     = buf[6];
   buflen  = buf[7];
   vdatap  = buf + 8;
   func    = VdataFunc;

   subtype = orechk[subindx].subtype;

   channel_code = (vdatap[2] >> 4) & 0x0f;
   channel = channelval(chstr, orechk[subindx].chsize, channel_code);

   if ( configp->oreid_16 == YES )
      vident = (unsigned short)buf[5] | (channel << 8);
   else
      vident = (unsigned short)buf[5];

   x10global.longvdata = 0;
   x10global.lastvtype = vtype;

   hcode = ucode = 0;
   found = 0;
   j = 0;
   /* Look up the alias, if any */
   idmask = configp->oreid_mask;
   while ( !found && aliasp && aliasp[j].line_no > 0 ) {
      if ( aliasp[j].vtype == vtype && aliasp[j].subtype == subtype && 
           (bitmap = aliasp[j].unitbmap) > 0 &&
           (ucode = single_bmap_unit(aliasp[j].unitbmap)) != 0xff ) {
         for ( k = 0; k < aliasp[j].nident; k++ ) {
            if ( (aliasp[j].ident[k] & idmask) == (vident & idmask) ) {
               index = j;
               found = 1;
               break;
            }
         }
      }
      j++;
   }

   if ( !found || !aliasp ) {
      if ( seq == 1 ) {
         sprintf(outbuf, "func %12s : Type %s_%s %sID 0x%02X Data 0x",
            "RFdata", typename[vtype], orechk[subindx].slabel, chstr, vident);
         for ( j = 0; j < buflen; j++ ) {
            sprintf(outbuf + strlen(outbuf), "%02x", vdatap[j]);
         }
      }
      return outbuf;
   }

   nvar = orechk[subindx].nvar;

   /* Internal check */
   if ( aliasp[index].storage_units != (2 * nvar) ) {
      sprintf(outbuf, "Internal error - storage != 2 * nvar\n");
      return outbuf;
   }
      
   
   hc = aliasp[index].housecode;
   hcode = hc2code(hc);
   unit = code2unit(ucode);

   /* Update activity timeout for sensor */
   aliasp[index].timeout = configp->inactive_timeout;

   /* Whether or not to display channel in output */
   if ( configp->ore_display_chan == NO )
      *chstr = '\0';

   switch ( subtype ) {
      case OreTH1 :
      case OreTH2 :
      case OreTH3 :
      case OreTH4 :
      case OreTH5 :
      case OreTH6 :
      case OreTemp1 :
      case OreTemp2 :
      case OreTHB1 :
      case OreTHB2 :
         loc = aliasp[index].storage_index;
         if ( seq == 1 ) {
            /* Save interval since last transmission */
            intvstrf(interval, time(NULL), x10state[hcode].timestamp[ucode]);

            /* Process temperature and battery data */
            for ( j = 0; j < nvar; j++ ) {
               prevdata[j] = x10global.data_storage[loc + j];
               lastdata[j] = x10global.data_storage[loc + nvar + j];
            }
            longvdata = (unsigned long)(channel & 0x0fu) << ORE_CHANSHFT;

            dtempc = 100 * ((vdatap[5] & 0xf0u) >> 4) +
                            10 * (vdatap[5] & 0x0fu) +
                                ((vdatap[4] & 0xf0u) >> 4);
            longvdata |= (unsigned long)dtempc << ORE_DATASHFT;
            status = ORE_VALID;

            if ( vdatap[6] & 0x08u ) {
               dtempc = -dtempc;
               status |= ORE_NEGTEMP;
            }

            *batstr = '\0';
            blevel = 0;

            switch ( orechk[subindx].batmode ) {
               case 2 :
                  blevel = 10 - (vdatap[4] & 0x0fu);
                  sprintf(batstr, "BatLvl %d%% ", blevel * 10);
                  status |= ORE_BATLVL;
                  if ( (10 * blevel) <= configp->ore_lobat ) {
                     status |= ORE_LOBAT;
                  }
                  break;
               case 1 :
                  if ( vdatap[4] & 0x04u ) {
                     status |= ORE_LOBAT;
                  }
                  break;
               default :
                  break;
            }
                     
            longvdata |= status | (blevel << ORE_BATSHFT);

            lastdtempc = (lastdata[0] & ORE_DATAMSK) >> ORE_DATASHFT;
            if ( lastdata[0] & ORE_NEGTEMP ) 
               lastdtempc = -lastdtempc;
            delta = abs(dtempc - lastdtempc);
            *sunchanged = (delta < (int)configp->ore_chgbits_t) ? 1 : 0;

            func = OreTempFunc;
            trig = OreTempTrig;
            vflags |= (status & ORE_LOBAT) ? SEC_LOBAT : 0;

            if ( (aliasp[index].optflags2 & MOPT2_TMIN) && (dtempc < aliasp[index].tmin) )
               vflags |= ORE_TMIN;
            if ( (aliasp[index].optflags2 & MOPT2_TMAX) && (dtempc > aliasp[index].tmax) )
               vflags |= ORE_TMAX;

            create_flagslist (vtype, vflags, flagslist);

            temp = celsius2temp((double)dtempc / 10.0, configp->ore_tscale, configp->ore_toffset);
            sprintf(outbuf, "func %12s : hu %c%-2d %sTemp "FMT_ORET"%c %s%s(%s)",  
               funclabel[func], hc, unit, chstr, temp, configp->ore_tscale,
               batstr, flagslist, aliasp[index].label);
         }
         else if ( seq == 2 ) {
            /* Process Humidity data */
            longvdata = (unsigned long)(channel & 0x0fu) << ORE_CHANSHFT;
            humid = 10 * (vdatap[7] & 0x0fu) + ((vdatap[6] & 0xf0u) >> 4);
            longvdata |= (unsigned long)humid << ORE_DATASHFT;
            status |= ORE_VALID;

            longvdata |= status | (blevel << 24);

            lasthumid = (lastdata[seq - 1] & ORE_DATAMSK) >> ORE_DATASHFT;             
            delta = abs(humid - lasthumid);
            *sunchanged = (delta < (int)configp->ore_chgbits_rh) ? 1 : 0;

            func = OreHumidFunc;
            trig = OreHumidTrig;
            vflags |= (status & ORE_LOBAT) ? SEC_LOBAT : 0;

            create_flagslist (vtype, vflags, flagslist);

            sprintf(outbuf, "func %12s : hu %c%-2d %sRH %d%% %s%s(%s)",
               funclabel[func], hc, unit, chstr, humid,
               batstr, flagslist, aliasp[index].label);
         }
         else if ( seq == 3 ) {
            /* Process Barometric Pressure data */
            longvdata = (unsigned long)(channel & 0x0fu) << ORE_CHANSHFT;
            baro = (int)vdatap[8] + orechk[subindx].bpoffset;
            longvdata |= (unsigned long)baro << ORE_DATASHFT;
            status |= ORE_VALID;

            longvdata |= status | (blevel << ORE_BATSHFT);

            lastbaro = (lastdata[seq - 1] & ORE_DATAMSK) >> ORE_DATASHFT;
            delta = abs(baro - lastbaro);
            *sunchanged = (delta < (int)configp->ore_chgbits_bp) ? 1 : 0;

            func = OreBaroFunc;
            trig = OreBaroTrig;
            vflags |= (status & ORE_LOBAT) ? SEC_LOBAT : 0;

            create_flagslist (vtype, vflags, flagslist);

            dbaro = (double)baro * configp->ore_bpscale + configp->ore_bpoffset;
            sprintf(outbuf, "func %12s : hu %c%-2d %sBP "FMT_OREBP"%s %s%s(%s)",
               funclabel[func], hc, unit, chstr, dbaro, configp->ore_bpunits,
               batstr, flagslist, aliasp[index].label);            
         }
         else {
            break;
         }

         x10global.longvdata = longvdata;

         x10global.data_storage[loc + seq - 1] = longvdata;
         x10state[hcode].longdata = longvdata;

         if ( *sunchanged == 0 ) {
            x10state[hcode].state[ChgState] |= bitmap;
            x10global.data_storage[loc + nvar + seq - 1] = longvdata;
         }
         else {
            x10state[hcode].state[ChgState] &= ~bitmap;
         }

#if 0
         sprintf(outbuf + strlen(outbuf), " %s", ((*sunchanged) ? "NOCHG":"CHG"));
#endif

         break;

      case OreWeight1 :         
         /* Save interval since last transmission */
         intvstrf(interval, time(NULL), x10state[hcode].timestamp[ucode]);

         /* Process weight data */
         loc = aliasp[index].storage_index;
         for ( j = 0; j < nvar; j++ ) {
            prevdata[j] = x10global.data_storage[loc + j];
            lastdata[j] = x10global.data_storage[loc + nvar + j];
         }

         longvdata = (unsigned long)(channel & 0x0fu) << ORE_CHANSHFT;

         /* Weight in units of 0.1 kg */
         cweight = 1000 * (vdatap[5] & 0x01u) + 
                    100 * ((vdatap[4] & 0xf0u) >> 4) +
                     10 * (vdatap[4] & 0x0fu) + 
                          ((vdatap[3] & 0xf0u) >> 4);

         /* These two bits cycle through 0-3 for each new weighing */
         wgtnum = (vdatap[2] & 0x30u) >> 3;

         /* Unknown bits in the transmission */
         uknbytes = ((vdatap[2] & ~0x30u) << 8) | (vdatap[3] & 0x0fu);

         longvdata |= (unsigned long)cweight << ORE_DATASHFT;
         status = wgtnum | ORE_VALID;

         *batstr = '\0';
         blevel = 0;

         longvdata |= status | (blevel << ORE_BATSHFT);

         lastcweight = (lastdata[0] & ORE_DATAMSK) >> ORE_DATASHFT;
         delta = abs(cweight - lastcweight);
         if ( (delta >= (int)configp->ore_chgbits_wgt) ||
              (wgtnum != (lastdata[0] & ORE_WGTNUM))    ||
              !(lastdata[0] & ORE_VALID)  ) {
            *sunchanged = 0;
         }
         else {
            *sunchanged = 1;
         }

         func = OreWeightFunc;
         trig = OreWeightTrig;
         vflags |= (status & ORE_LOBAT) ? SEC_LOBAT : 0;

         create_flagslist (vtype, vflags, flagslist);

         weight = (double)cweight / 10.0 * configp->ore_wgtscale;
         sprintf(outbuf, "func %12s : hu %c%-2d %sWgt "FMT_OREWGT"%s %s%s(%s) ub = 0x%04x",
            funclabel[func], hc, unit, chstr, weight, configp->ore_wgtunits,
            batstr, flagslist, aliasp[index].label, uknbytes);

         x10global.longvdata = longvdata;

         x10global.data_storage[loc + seq - 1] = longvdata;
         x10state[hcode].longdata = longvdata;

         if ( *sunchanged == 0 ) {
            x10state[hcode].state[ChgState] |= bitmap;
            x10global.data_storage[loc + nvar + seq - 1] = longvdata;
         }
         else {
            x10state[hcode].state[ChgState] &= ~bitmap;
         }

#if 0
     sprintf(outbuf + strlen(outbuf), " %s", ((*sunchanged) ? "NOCHG":"CHG"));
#endif

         break;

      default :
         break;
   }

   if ( configp->display_sensor_intv == YES && (aliasp[index].optflags & MOPT_HEARTBEAT) )
      snprintf(outbuf + strlen(outbuf), sizeof(outbuf), " Intv %s ", interval);

   x10state[hcode].vaddress = bitmap;
   x10state[hcode].lastcmd = func;
   x10state[hcode].lastunit = unit;
   x10state[hcode].vident[ucode] = vident;
   x10state[hcode].vflags[ucode] = vflags;
   x10state[hcode].timestamp[ucode] = time(NULL);
   x10global.lasthc = hcode;
   x10global.lastaddr = 0;


   actfunc = 0;
   afuncmap = 0;
   xactfunc = func;
   xfuncmap = (1 << trig);
   trigaddr = bitmap;

   mask = modmask[VdataMask][hcode];
   active = bitmap & mask;

   if ( i_am_state )
      write_x10state_file();

   /* Heyuhelper, if applicable */
   if ( i_am_state && signal_source & RCVI && configp->script_mode & HEYU_HELPER ) {
      launch_heyuhelper(hcode, trigaddr, func);
      return 0;
   }

   bmaplaunch = 0;
   j = 0;
   while ( launcherp && launcherp[j].line_no > 0 ) {
      if ( (launcherp[j].flags & x10global.flags) != launcherp[j].flags ||
	   (launcherp[j].notflags & ~x10global.flags) != launcherp[j].notflags ||
           (launcherp[j].vflags & x10state[hcode].vflags[ucode]) != launcherp[j].vflags ||
	   (launcherp[j].notvflags & ~x10state[hcode].vflags[ucode]) != launcherp[j].notvflags ||
           (launcherp[j].czflags & x10global.czflags) != launcherp[j].czflags ||
	   (launcherp[j].notczflags & ~x10global.czflags) != launcherp[j].notczflags ||
            launcherp[j].type != L_NORMAL ) {
         j++;
	 continue;
      }

      if ( launcherp[j].hcode == hcode &&
           launcherp[j].xfuncmap & xfuncmap &&
           launcherp[j].source & signal_source ) {
         trigactive = trigaddr & (mask | launcherp[j].signal);
         if ( (bmaplaunch = (trigactive & launcherp[j].bmaptrig) & 
                   (x10state[hcode].state[ChgState] | ~launcherp[j].chgtrig)) ||
                   (launcherp[j].unitstar && !trigaddr)) {
            *launchp = j;
            launcherp[j].actfunc = xactfunc;
            launcherp[j].genfunc = 0;
	    launcherp[j].xfunc = xactfunc;
	    launcherp[j].level = x10state[hcode].dimlevel[ucode];
            launcherp[j].bmaplaunch = bmaplaunch;
            launcherp[j].actsource = signal_source;
            break;
         }
      }
      j++;
   }

   x10state[hcode].launched = bmaplaunch;

   return outbuf;
#else
   return "";
#endif /* HASORE */
}   

/*----------------------------------------------------------------------------+
 | Translate Oregon RF data                                                   |
 | Buffer reference: ST_COMMAND, ST_ORE_EMU, plus:                            |
 |  vtype, subindx, hcode, ucode, seq, data[0], data[1], data[2], data[3]     |
 |   [2]     [3]     [4]    [5]   [6]    [7]     [8]      [9]      [10]       |
 +----------------------------------------------------------------------------*/
char *translate_ore_emu( unsigned char *buf, unsigned char *sunchanged, int *launchp )
{
#ifdef HASORE
   static char    outbuf[160];
   static char    interval[32];
   char           flagslist[80];
   ALIAS          *aliasp;
   LAUNCHER       *launcherp;
   extern unsigned int signal_source;
   unsigned char  func, vtype, subindx, subtype, seq;
   int            humid, lasthumid;
   int            baro, lastbaro;
   unsigned short vident;
   int            unit, loc;
   int            dtempc, lastdtempc, delta, nvar;
   unsigned long  longvdata;

   char           hc;
   char           chstr[8];

   unsigned char  actfunc, xactfunc;
   unsigned int   trigaddr, mask, active, trigactive;
   unsigned int   bmaplaunch;
   unsigned long  afuncmap, xfuncmap;

   int            j, /*k,*/ found, index = -1;
   unsigned char  hcode, ucode, trig = 0;
   unsigned int   bitmap;
   unsigned long  vflags = 0;

   unsigned long prevdata, lastdata;

   double temp, dbaro;

   static char *orecmd[] = {"?", "oretemp", "orerh", "orebp"};

   launcherp = configp->launcherp;
   aliasp = configp->aliasp;

   *launchp = -1;
   *sunchanged = 0;
   flagslist[0] = '\0';
   *outbuf = '\0';


   vtype   = buf[2];
   subindx = buf[3];
   hcode   = buf[4];
   ucode   = buf[5];
   seq     = buf[6];
   func    = VdataFunc;

   subtype = orechk[subindx].subtype;

   x10global.longvdata = 0;
   x10global.lastvtype = vtype;

   hc = code2hc(hcode);
   bitmap = 1 << ucode;
   unit = code2unit(ucode);

   /* Reconstruct unsigned long from littleendian bytes */
   longvdata = 0;
   for ( j = 0; j < 4; j++ ) 
      longvdata |= buf[7 + j] << (8 * j);

   /* Whether or not to display channel in output */
   if ( configp->ore_display_chan == YES )
      sprintf(chstr, "Ch %d ", (int)(longvdata & ORE_CHANMSK) >> ORE_CHANSHFT);
   else
      *chstr = '\0';
         
   found = 0;
   j = 0;
   /* Look up the alias, if any */
   while ( !found && aliasp && aliasp[j].line_no > 0 ) {
      if ( aliasp[j].vtype == vtype && aliasp[j].subtype == subtype && 
           bitmap == aliasp[j].unitbmap && hc == aliasp[j].housecode ) {
         index = j;
         found = 1;
         break;
      }
      j++;
   }

   if ( !found || !aliasp ) {
      sprintf(outbuf, "func %12s : hu %c%-2d func %s %sdata %lu - Missing Alias",
         "oreEmu", hc, unit, orecmd[seq], chstr, longvdata);
      return outbuf;
   }

   /* Save interval since last transmission */
   intvstrf(interval, time(NULL), x10state[hcode].timestamp[ucode]);

   vident = aliasp[index].ident[0];

   nvar = orechk[subindx].nvar;

   /* Internal check */
   if ( aliasp[index].storage_units != (2 * nvar) ) {
      sprintf(outbuf, "Internal error - storage != 2 * nvar\n");
      return outbuf;
   }
      
   /* Update activity timeout for sensor */
   aliasp[index].timeout = configp->inactive_timeout;

   loc = aliasp[index].storage_index;

   prevdata = x10global.data_storage[loc + seq - 1];
   lastdata = x10global.data_storage[loc + nvar + seq - 1];

   switch ( seq ) {
      case 1 :  /* Temperature */
         dtempc = (longvdata & ORE_DATAMSK) >> ORE_DATASHFT;
         if ( longvdata & ORE_NEGTEMP )
            dtempc = -dtempc;

         lastdtempc = (lastdata & ORE_DATAMSK) >> ORE_DATASHFT;
         if ( lastdata & ORE_NEGTEMP ) 
            lastdtempc = -lastdtempc;
         delta = abs(dtempc - lastdtempc);
         *sunchanged = (delta < (int)configp->ore_chgbits_t) ? 1 : 0;

         func = OreTempFunc;
         trig = OreTempTrig;
         vflags = 0;

         if ( (aliasp[index].optflags2 & MOPT2_TMIN) && (dtempc < aliasp[index].tmin) )
            vflags |= ORE_TMIN;
         if ( (aliasp[index].optflags2 & MOPT2_TMAX) && (dtempc > aliasp[index].tmax) )
            vflags |= ORE_TMAX;

         create_flagslist (vtype, vflags, flagslist);

         temp = celsius2temp((double)dtempc / 10.0, configp->ore_tscale, configp->ore_toffset);
         sprintf(outbuf, "func %12s : hu %c%-2d %sTemp "FMT_ORET"%c %s(%s)",  
            funclabel[func], hc, unit, chstr, temp, configp->ore_tscale,
            flagslist, aliasp[index].label);
         break;

      case 2 : /* Relative Humidity */
         humid = (longvdata & ORE_DATAMSK) >> ORE_DATASHFT;
         lasthumid = (lastdata & ORE_DATAMSK) >> ORE_DATASHFT;
         delta = abs(humid - lasthumid);
         *sunchanged = (delta < (int)configp->ore_chgbits_rh) ? 1 : 0;

         func = OreHumidFunc;
         trig = OreHumidTrig;
         vflags = 0;

         create_flagslist (vtype, vflags, flagslist);

         sprintf(outbuf, "func %12s : hu %c%-2d %sRH %d%% %s(%s)",
            funclabel[func], hc, unit, chstr, humid, flagslist, aliasp[index].label);
         break;

      case 3 : /* Barometric Pressure */
         baro = (int)((longvdata & ORE_DATAMSK) >> ORE_DATASHFT);
         longvdata = (longvdata & ~ORE_DATAMSK) | ((unsigned long)baro << ORE_DATASHFT);
         lastbaro = (lastdata & ORE_DATAMSK) >> ORE_DATASHFT;
         delta = abs(baro - lastbaro);
         *sunchanged = (delta < (int)configp->ore_chgbits_bp) ? 1 : 0;

         func = OreBaroFunc;
         trig = OreBaroTrig;
         vflags = 0;

         create_flagslist (vtype, vflags, flagslist);

         dbaro = (double)baro * configp->ore_bpscale + configp->ore_bpoffset;
         sprintf(outbuf, "func %12s : hu %c%-2d %sBP "FMT_OREBP"%s %s(%s)",
            funclabel[func], hc, unit, chstr, dbaro, configp->ore_bpunits,
            flagslist, aliasp[index].label);
         break;

      default :
         break;
   }            

   if ( configp->display_sensor_intv == YES && (aliasp[index].optflags & MOPT_HEARTBEAT) )
      snprintf(outbuf + strlen(outbuf), sizeof(outbuf), " Intv %s ", interval);

#if 0
         sprintf(outbuf + strlen(outbuf), " %s", ((*sunchanged) ? "NOCHG":"CHG"));
#endif


   x10global.longvdata = longvdata;

   x10global.data_storage[loc + seq - 1] = longvdata;
   x10state[hcode].longdata = longvdata;

   if ( *sunchanged == 0 ) {
      x10state[hcode].state[ChgState] |= bitmap;
      x10global.data_storage[loc + nvar + seq - 1] = longvdata;
   }
   else {
      x10state[hcode].state[ChgState] &= ~bitmap;
   }

#if 0
         sprintf(outbuf + strlen(outbuf), " %s", ((*sunchanged) ? "NOCHG":"CHG"));
#endif


   x10state[hcode].vaddress = bitmap;
   x10state[hcode].lastcmd = func;
   x10state[hcode].lastunit = unit;
   x10state[hcode].vident[ucode] = vident;
   x10state[hcode].vflags[ucode] = vflags;
   x10state[hcode].timestamp[ucode] = time(NULL);
   x10global.lasthc = hcode;
   x10global.lastaddr = 0;


   actfunc = 0;
   afuncmap = 0;
   xactfunc = func;
   xfuncmap = (1 << trig);
   trigaddr = bitmap;

   mask = modmask[VdataMask][hcode];
   active = bitmap & mask;

   if ( i_am_state )
      write_x10state_file();

   /* Heyuhelper, if applicable */
   if ( i_am_state && signal_source & RCVI && configp->script_mode & HEYU_HELPER ) {
      launch_heyuhelper(hcode, trigaddr, func);
      return 0;
   }

   bmaplaunch = 0;
   j = 0;
   while ( launcherp && launcherp[j].line_no > 0 ) {
      if ( (launcherp[j].flags & x10global.flags) != launcherp[j].flags ||
	   (launcherp[j].notflags & ~x10global.flags) != launcherp[j].notflags ||
           (launcherp[j].vflags & x10state[hcode].vflags[ucode]) != launcherp[j].vflags ||
	   (launcherp[j].notvflags & ~x10state[hcode].vflags[ucode]) != launcherp[j].notvflags ||
           (launcherp[j].czflags & x10global.czflags) != launcherp[j].czflags ||
	   (launcherp[j].notczflags & ~x10global.czflags) != launcherp[j].notczflags ||
            launcherp[j].type != L_NORMAL ) {
         j++;
	 continue;
      }

      if ( launcherp[j].hcode == hcode &&
           launcherp[j].xfuncmap & xfuncmap &&
           launcherp[j].source & signal_source ) {
         trigactive = trigaddr & (mask | launcherp[j].signal);
         if ( (bmaplaunch = (trigactive & launcherp[j].bmaptrig) & 
                   (x10state[hcode].state[ChgState] | ~launcherp[j].chgtrig)) ||
                   (launcherp[j].unitstar && !trigaddr)) {
            *launchp = j;
            launcherp[j].actfunc = xactfunc;
            launcherp[j].genfunc = 0;
	    launcherp[j].xfunc = xactfunc;
	    launcherp[j].level = x10state[hcode].dimlevel[ucode];
            launcherp[j].bmaplaunch = bmaplaunch;
            launcherp[j].actsource = signal_source;
            break;
         }
      }
      j++;
   }

   x10state[hcode].launched = bmaplaunch;

   return outbuf;
#else
   return "";
#endif /* HASORE */
} 


#ifdef HASORE
/*---------------------------------------------------------------------+
 | Display stored data for Oregon sensors                              |
 +---------------------------------------------------------------------*/
int show_oregon ( void )
{

   ALIAS          *aliasp;
   char           hc;
   int            j, unit, index, count = 0, maxlabel = 0, maxmod = 0;
   int            loc, locunits;
   unsigned char  hcode, ucode, func;
   unsigned char  subtype, subindx, status = 0;
   unsigned long  longvdata = 0;
   short          dtempc, cweight;
   int            baro, blevel;
   int            channel;
   double         temp, dbaro, weight;

   if ( !(aliasp = configp->aliasp) ) 
      return 0;

   /* Get maximum lengths of module name and alias label */
   index = 0;
   while ( aliasp[index].line_no > 0 ) {
      if ( aliasp[index].vtype == RF_OREGON ) {
         count++;
         maxlabel = max(maxlabel, (int)strlen(aliasp[index].label));
         maxmod = max(maxmod, (int)strlen(lookup_module_name(aliasp[index].modtype)) );
      }
      index++;
   }

   if ( count == 0 )
      return 0;

   index = 0;
   while ( aliasp[index].line_no > 0 ) {
      if ( aliasp[index].vtype != RF_OREGON ) {
         index++;
         continue;
      }
      ucode = single_bmap_unit(aliasp[index].unitbmap);
      unit = code2unit(ucode);
      hc = aliasp[index].housecode;
      hcode = hc2code(hc);
      subtype = aliasp[index].subtype;
      loc = aliasp[index].storage_index;
      locunits = aliasp[index].storage_units;

      for ( subindx = 0; subindx < orechk_size; subindx++ ) {
         if ( orechk[subindx].subtype == subtype )
            break;
      }

      printf("%c%-2d %-*s %-*s ", hc, unit,
          maxmod, lookup_module_name(aliasp[index].modtype),
          maxlabel, aliasp[index].label);

      channel = 0;
      for ( j = 0; j < locunits; j++ ) {
          longvdata = x10global.data_storage[loc + j];
          if ( longvdata & ORE_VALID )
             channel = (int)((longvdata & ORE_CHANMSK) >> ORE_CHANSHFT);
      }
      if ( configp->ore_display_chan == YES ) {
         if ( channel > 0 )
            printf("Ch %d ", channel);
         else
            printf("Ch - ");
      }

      for ( j = 0; j < locunits; j++ ) {
         func = aliasp[index].funclist[j];
         longvdata = x10global.data_storage[loc + j];
         status = longvdata & 0xff;
         switch ( func ) {
            case OreTempFunc :
               if ( !(longvdata & ORE_VALID) ) {
                  printf("Temp ----%c ", configp->ore_tscale);
                  break;
               }

               dtempc = (longvdata & ORE_DATAMSK) >> ORE_DATASHFT;
               if ( status & ORE_NEGTEMP )
                  dtempc = -dtempc;
               temp = celsius2temp((double)dtempc / 10.0, configp->ore_tscale, configp->ore_toffset);
               printf("Temp "FMT_ORET"%c ", temp, configp->ore_tscale);
               break;
            case OreHumidFunc :
               if ( !(longvdata & ORE_VALID) ) {
                  printf("RH --%% ");
                  break;
               }
               printf("RH %2d%% ", (int)(longvdata & ORE_DATAMSK) >> ORE_DATASHFT);
               break;
            case OreBaroFunc :
               if ( !(longvdata & ORE_VALID) ) {
                  printf("BP ---%s ", configp->ore_bpunits);
                  break;
               }
               baro = (int)(((longvdata & ORE_DATAMSK) >> ORE_DATASHFT) );
               dbaro = (double)baro * configp->ore_bpscale + configp->ore_bpoffset;
               printf("BP "FMT_OREBP"%s ", dbaro, configp->ore_bpunits);
               break;
            case OreWeightFunc :
               if ( !(longvdata & ORE_VALID) ) {
                  printf("Weight ---%s ", configp->ore_wgtunits);
                  break;
               }
               cweight = (longvdata & ORE_DATAMSK) >> ORE_DATASHFT;
               weight = ((double)cweight / 10.0) * configp->ore_wgtscale;
               printf("Wgt "FMT_OREWGT"%s ", weight, configp->ore_wgtunits);
               break;
         }
      }

      if ( status & ORE_BATLVL ) {
         blevel = 10 * (int)((longvdata & ORE_BATMSK) >> ORE_BATSHFT);
         printf("BatLvl %d%% ", blevel );
         if ( blevel <= configp->ore_lobat )
            printf("LoBat ");
      }
      else if ( status & ORE_LOBAT ) {
         printf("LoBat ");
      }

      printf("\n");
      index++;
   }

   return 0;
}
#endif /* HASORE */

#ifdef HASORE 
/*---------------------------------------------------------------------+
 | Display an Oregon data value stored in the x10state structure.      |
 +---------------------------------------------------------------------*/
int c_orecmds ( int argc, char *argv[] )
{

   ALIAS          *aliasp;
   unsigned char  hcode, ucode, func;
   unsigned long  aflags;
   unsigned long  longvdata;
   char           hc;
   unsigned int   bitmap;
   int            j;
   int            unit, index, dtempc, baro, cweight, loc;
   double         temp, dbaro, weight;

   static struct {
      char          *cmd;
      unsigned char func;
   } cmdlist[] = {
      {"oretemp",   OreTempFunc   },
      {"orerh",     OreHumidFunc  },
      {"orebp",     OreBaroFunc   },
      {"orewgt",    OreWeightFunc },
      {"",          0xff          },
   };

   int read_x10state_file ( void );

   if ( argc < 3 ) {
      fprintf(stderr, "Usage: %s %s Hu\n", argv[0], argv[1]);
      return 1;
   }

   if ( check_for_engine() != 0 ) {
      fprintf(stderr, "State engine is not running.\n");
      return 1;
   }
   if ( read_x10state_file() != 0 ) {
      fprintf(stderr, "Unable to read state file.\n");
      return 1;
   }

   if ( (aliasp = configp->aliasp) == NULL )
      return 1;

   j = 0;
   func = 0xff;
   while ( *cmdlist[j].cmd ) {
      if ( strcmp(argv[1], cmdlist[j].cmd) == 0 ) {
         func = cmdlist[j].func;
         break;
      }
      j++;
   }
      
   aflags = parse_addr(argv[2], &hc, &bitmap);
   if ( !(aflags & A_VALID) || aflags & (A_DUMMY | A_PLUS | A_MINUS) || bitmap == 0 ) {
      fprintf(stderr, "Invalid Hu address '%s'\n", argv[2]);
      return 1;
   }
   if ( aflags & A_MULT ) {
      fprintf(stderr, "Only a single unit address is valid.\n");
      return 1;
   }
   hcode = hc2code(hc);
   ucode = single_bmap_unit(bitmap);
   unit  = code2unit(ucode);

   if ( (index = alias_lookup_index(hc, bitmap, RF_OREGON)) < 0 ) {
      fprintf(stderr, "Address %c%d is not configured as an Oregon sensor\n", hc, unit);
      return 1;
   }

   loc = -1;
   for ( j = 0; j < aliasp[index].storage_units; j++ ) {
      if ( aliasp[index].funclist[j] == func ) {
         loc = aliasp[index].storage_index + j;
         break;
      }
   }
   if ( loc < 0 ) {
      fprintf(stderr, "Command '%s' not valid for this sensor.\n", argv[1]);
      return 1;
   }

   longvdata = x10global.data_storage[loc];

   if ( !(longvdata & ORE_VALID) ) {
      fprintf(stderr, "Not ready\n");
      return 1;
   }

   switch ( func ) {
      case OreTempFunc :
         dtempc = (longvdata & ORE_DATAMSK) >> ORE_DATASHFT;
         if ( longvdata & ORE_NEGTEMP )
            dtempc = -dtempc;
         temp = celsius2temp((double)dtempc / 10.0, configp->ore_tscale, configp->ore_toffset);
         printf("%.1f\n", temp);
         break;
     case OreHumidFunc :
         printf("%d\n", (int)((longvdata & ORE_DATAMSK) >> ORE_DATASHFT));
         break;
     case OreBaroFunc :
         baro = (longvdata & ORE_DATAMSK) >> ORE_DATASHFT;
         dbaro = (double)baro * configp->ore_bpscale + configp->ore_bpoffset;
         printf("%.1f\n", dbaro);
         break;
     case OreWeightFunc :
         cweight = (longvdata & ORE_DATAMSK) >> ORE_DATASHFT;
         weight = (double)cweight / 10.0 * configp->ore_wgtscale;
         printf(FMT_OREWGT"\n", weight);
         break;
     default :
         break;
  }
  return 0;
}




/*----------------------------------------------------------------------------+
 | Send emulated Oregon data to dummy module.                                 |
 | Buffer reference: ST_COMMAND, ST_ORE_EMU, plus:                            |
 |  vtype, subindex, hcode, ucode, seq, data[0], data[1], data[2], data[3]    |
 |   [2]     [3]      [4]    [5]   [6]    [7]     [8]      [9]      [10]      |
 | where data[0] ... data[3] are the littleendian bytes of a 4-byte unsigned  |
 | long value.                                                                |
 +----------------------------------------------------------------------------*/
int send_ore_emu ( unsigned char vtype, unsigned char subindex, int seq,
    unsigned char hcode, unsigned char ucode, unsigned long lvalue )
{
   extern int sptty;
   extern int heyu_parent;
   char writefilename[PATH_LEN + 1];
   unsigned char outbuf[128];
   int  j;
   int  ignoret;

   static unsigned char template[18] = {
    0xff,0xff,0xff,3,ST_COMMAND,ST_SOURCE,0,
    0xff,0xff,0xff,0,ST_COMMAND,ST_ORE_EMU,0,0,0,0,0};

   sprintf(writefilename, "%s%s", WRITEFILE, configp->suffix);

   template[6]  = (unsigned char)heyu_parent;
   template[10] = 11;
   template[13] = vtype;
   template[14] = subindex;
   template[15] = hcode;
   template[16] = ucode;
   template[17] = seq;

   memcpy(outbuf, template, sizeof(template));
   for ( j = 0; j < 4; j++ ) {
      outbuf[(sizeof(template) + j)] = (lvalue >> (8 * j)) & 0xff;
   }

   ignoret = write(sptty, (char *)outbuf, (sizeof(template)) + 4);

   return 0;
}

/*---------------------------------------------------------------------+
 | Store an emulated Oregon data value in the x10state structure.      |
 | 'heyu ore_emu Hu <orefunc> <funcvalue>'                             |
 +---------------------------------------------------------------------*/
int c_ore_emu ( int argc, char *argv[] )
{

   ALIAS          *aliasp;
   unsigned char  hcode, ucode, func, seq, subtype;
   unsigned long  aflags;
   unsigned long  longvdata;
   char           hc;
   char           tscale;
   unsigned int   bitmap;
   int            j, subindex;
   int            unit, index,loc;
   double         tempc, dvalue;
   unsigned char  *dunit;
   char           bpunits[sizeof(config.ore_bpunits)];
   char           bpinunits[sizeof(config.ore_bpunits)];

   int            channel = 1;  /* Dummy channel */

   static struct {
      char          *cmd;
      unsigned char func;
   } cmdlist[] = {
      {"oretemp",   OreTempFunc   },
      {"orerh",     OreHumidFunc  },
      {"orebp",     OreBaroFunc   },
      {"",          0xff          },
   };

   int read_x10state_file ( void );

   if ( argc < 5 ) {
      fprintf(stderr, "Usage: %s %s Hu <oretemp|orerh|orebp> <value>\n", argv[0], argv[1]);
      return 1;
   }

   if ( check_for_engine() != 0 ) {
      fprintf(stderr, "State engine is not running.\n");
      return 1;
   }
   if ( read_x10state_file() != 0 ) {
      fprintf(stderr, "Unable to read state file.\n");
      return 1;
   }

   if ( (aliasp = configp->aliasp) == NULL )
      return 1;

   j = 0;
   func = 0xff;
   while ( *cmdlist[j].cmd ) {
      if ( strcmp(argv[3], cmdlist[j].cmd) == 0 ) {
         func = cmdlist[j].func;
         break;
      }
      j++;
   }

   if ( func == 0xff ) {
      fprintf(stderr, "Function must be oretemp, orerh, or orebp.\n");
      return 1;
   }
      
   aflags = parse_addr(argv[2], &hc, &bitmap);
   if ( !(aflags & A_VALID) || aflags & (A_DUMMY | A_PLUS | A_MINUS) || bitmap == 0 ) {
      fprintf(stderr, "Invalid Hu address '%s'\n", argv[2]);
      return 1;
   }
   if ( aflags & A_MULT ) {
      fprintf(stderr, "Only a single unit address is valid.\n");
      return 1;
   }
   hcode = hc2code(hc);
   ucode = single_bmap_unit(bitmap);
   unit  = code2unit(ucode);

   if ( (index = alias_lookup_index(hc, bitmap, RF_OREGON)) < 0 ) {
      fprintf(stderr, "Address %c%d is not configured as an Oregon emulation module type\n", hc, unit);
      return 1;
   }

   subtype = aliasp[index].subtype;
   subindex = -1;
   for ( j = 0; j < orechk_size; j++ ) {
      if ( orechk[j].subtype == subtype && orechk[j].support_mode == 2 ) {
         subindex = j;
         break;
      }
   }

   if ( subindex < 0 ) {
      fprintf(stderr, "Module is not an Oregon emulation dummy type\n");
      return 1;
   }
      
   loc = -1;
   for ( j = 0; j < aliasp[index].storage_units; j++ ) {
      if ( aliasp[index].funclist[j] == func ) {
         loc = aliasp[index].storage_index + j;
         break;
      }
   }
   if ( loc < 0 ) {
      fprintf(stderr, "Function '%s' is not supported for this sensor.\n", argv[3]);
      return 1;
   }

   tscale = (configp->ore_data_entry == SCALED) ? configp->ore_tscale : 'C';

   dvalue = strtod(argv[4], (char **)(&dunit));

   longvdata = 0;
   switch ( func ) {
      case OreTempFunc :
         if ( !strchr(" /t/r/nCFKR", toupper(*dunit)) ) {
            fprintf(stderr, "Invalid temperature scale '%s'\n", dunit);
            return 1;
         }
         if ( *dunit && strchr("CFKR", toupper(*dunit)) )
            tscale = toupper(*dunit);

         tempc = temp2celsius(dvalue, tscale, 0.0);
          
         if ( tempc < 0.0 ) {
            longvdata |= ORE_NEGTEMP;
            tempc = -tempc;
         }
         if ( tempc > 99.0 ) {
            fprintf(stderr, "Temperature "FMT_ORET" is out of range.\n", dvalue);
            return 1;
         }
         longvdata |= ((unsigned long)((tempc + .05) * 10.0)) << 8 | ORE_VALID;
         longvdata |= (unsigned long)channel << ORE_CHANSHFT;
         seq = 1;
         break;
     case OreHumidFunc :
         if ( orechk[subindex].nvar < 2 ) {
            fprintf(stderr, "Function orerh is invalid for this module type.\n");
            return 1;
         }

         if ( !strchr(" /t/r/n%%", *dunit) ) {
            fprintf(stderr, "Invalid RH scale '%c'\n", *dunit);
            return 1;
         }
         if ( dvalue < 0.0 || dvalue > 255.0 ) {
            fprintf(stderr, "Relative Humidity %.0f%% is out of range.\n", dvalue);
            return 1;
         }
         longvdata = ((unsigned long)(dvalue + 0.5)) << 8 | ORE_VALID;
         longvdata |= (unsigned long)channel << ORE_CHANSHFT;
         seq = 2;
         break;
     case OreBaroFunc :
         if ( orechk[subindex].nvar < 3 ) {
            fprintf(stderr, "Function orebp is invalid for this module type.\n");
            return 1;
         }
         if ( !strchr(" /t/r/n", *dunit) ) {
            strncpy2(bpinunits, (char *)dunit, sizeof(bpinunits) - 1);
            strlower(bpinunits);
            strncpy2(bpunits, configp->ore_bpunits, sizeof(bpunits) - 1);
            strlower(bpunits);
            if ( strcmp(bpinunits, bpunits) == 0 ) {
               longvdata = (unsigned long)(dvalue / configp->ore_bpscale);
            }
            else if ( strcmp(bpinunits, "hpa") == 0 ) {
               longvdata = (unsigned long)dvalue;
            }
            else {
               fprintf(stderr, "BP unit '%s' not recognized\n", dunit);
               return 1;
            }
         }
         else if ( configp->ore_data_entry == SCALED ) {
            longvdata = (unsigned long)(dvalue /configp->ore_bpscale);
         }
         else {
            longvdata = (unsigned long)dvalue;
         }

         if ( dvalue < 0.0 || longvdata > 0xffffu ) {
            fprintf(stderr, "Barometric Pressure "FMT_OREBP" is out of range.\n", dvalue);
            return 1;
         }
         longvdata = (longvdata << ORE_DATASHFT) | ORE_VALID;
         longvdata |= (unsigned long)channel << ORE_CHANSHFT;
         seq = 3;
         break;

     default :
         return 0;
         break;
  }

  send_ore_emu(RF_OREGON, subindex, seq, hcode, ucode, longvdata);

  return 0;
}


#endif /* HASORE */

                          
int ore_maxmin_temp ( ALIAS *aliasp, int aliasindex, char **tokens, int *ntokens )
{
#ifdef HASORE
   double tvalue;
   int    tflag;
   char   tscale, tscale_init;

   int temp_parm ( char **, int *, char *, char *, int *, double * );

   if ( configp->ore_data_entry == NATIVE )
      tscale_init = 'C';
   else
      tscale_init = configp->ore_tscale;

   tscale = tscale_init;
   if ( temp_parm(tokens, ntokens, "TMIN", &tscale, &tflag, &tvalue) != 0 )
      return 1;
   if ( tflag ) {
      aliasp[aliasindex].optflags2 |= MOPT2_TMIN;
      tvalue = temp2celsius (tvalue, tscale, 0.0 );
      if ( tvalue < -200.0 || tvalue > 200.0 ) {
         store_error_message("TMIN value is out of range.");
         return 1;
      }

      /* Round to nearest 0.1C */
      if ( tvalue < 0.0 )
         tvalue = (tvalue * 10.0) - 0.5;
      else
         tvalue = (tvalue * 10.0) + 0.5;

      aliasp[aliasindex].tmin = (int)tvalue;
   }

   tscale = tscale_init;
   if ( temp_parm(tokens, ntokens, "TMAX", &tscale, &tflag, &tvalue) != 0 )
      return 1;
   if ( tflag ) {
      aliasp[aliasindex].optflags2 |= MOPT2_TMAX;
      tvalue = temp2celsius (tvalue, tscale, 0.0 );
      if ( tvalue < -200.0 || tvalue > 200.0 ) {
         store_error_message("TMAX value is out of range.");
         return 1;
      }

      /* Round to nearest 0.1C */
      if ( tvalue < 0.0 )
         tvalue = (tvalue * 10.0) - 0.5;
      else
         tvalue = (tvalue * 10.0) + 0.5;

      aliasp[aliasindex].tmax = (int)tvalue;
   }

   return 0;
#else
   return 0;
#endif
}   

